# This program is in the public domain
# Author: Paul Kienzle
"""
Refl1D: Specular Reflectometry Modeling and Fitting Software

This package provides tools for modeling a variety of systems from
simple slabs to complex systems with a mixture models involving
smoothly varying layers.

X-ray and neutron and polarized neutron data can be loaded and the model
parameters adjusted to achieve the best fit.

A graphical interface allows direct manipulation of the model profiles.

See http://www.reflectometry.org/danse/reflectometry for online manuals.
"""

__version__ = "0.8.10"
